------------------------------------------------------------------------
-- Implementation of Melissa O'Neill's variant of the prime sieve.
-- I only implement the bounded case and I don't implement wheels yet.
-- This is also not proven correct, you'll have to take my word for it
-- that it does what I claim.
------------------------------------------------------------------------
{-# OPTIONS --safe --without-K #-}

module Eratosthenes where

open import Data.Nat.Base
open import Data.Nat.Induction using (<-wellFounded-fast)
open import Data.Nat.Properties hiding (≤-total; ≤-isTotalOrder; ≤-totalOrder)
open import Data.List.Base hiding (upTo)
open import Data.Product.Base
open import Data.Sum.Base using (inj₁; inj₂)
open import Function.Base
open import Induction.WellFounded
import Relation.Binary.Construct.On as On
open import Relation.Binary.Bundles using (TotalOrder)
open import Relation.Binary.Definitions using (Total)
open import Relation.Binary.Structures using (IsTotalOrder)
open import Relation.Binary.PropositionalEquality
open import Relation.Nullary.Construct.Add.Extrema
open import Relation.Nullary.Decidable

-- Reimplementations of a couple of things from stdlib because
-- their existing definitions at time of writing are slow

-- ≤-total is currently defined in stdlib using unary arithmetic. This makes it
-- terrible to use as a conditional. Our heap implementation is generic over a
-- total order, so we redefine this and the bundle we care ultimately care
-- about.
≤-total : Total _≤_
≤-total m n with m ≤? n
... | yes m≤n = inj₁ m≤n
... | no m≰n = inj₂ (≰⇒≥ m≰n)

≤-isTotalOrder : IsTotalOrder _≡_ _≤_
≤-isTotalOrder = record
  { isPartialOrder = ≤-isPartialOrder
  ; total = ≤-total
  }

≤-totalOrder : TotalOrder _ _ _
≤-totalOrder = record { isTotalOrder = ≤-isTotalOrder }

-- upTo in stdlib creates larger and larger closures. This causes some slowdown.
-- We implement it in a tail recursive manner instead.
upFromThen : ℕ → ℕ → List ℕ
upFromThen from zero = []
upFromThen from (suc then) = from ∷ upFromThen (suc from) then

upTo : ℕ → List ℕ
upTo = upFromThen 0

open import SplayHeap (On.totalOrder ≤-totalOrder (proj₂ {A = ℕ}))

insertPrime : ℕ → Heap → Heap
insertPrime p table = insert (p , p * p) table

sieve′ : List ℕ → (t : Heap) → .{{NonEmpty t}} → List ℕ
sieve′ [] table = []
sieve′ (x ∷ xs) table with (p , nextComposite) , table′ ← findDeleteMin table with nextComposite ≤? x
... | no _ = x ∷ sieve′ xs (insertPrime x table)
... | yes _ = sieve′ xs adjusted
  where
    P : Heap → Set
    P _ = List (ℕ × ℕ) × Heap

    adjust-rec : ∀ (t : Heap) → (∀ {t′ : Heap} → # t′ < # t → P t′) → P t
    adjust-rec t rec with nonEmpty? t
    ... | no _ = [] , empty
    ... | yes ne with (p , c) , t′ ← findDeleteMin t {{ne}} in eq with c ≤? x
    ... | no _ = [] , t
    ... | yes _ = map₁ ((p , c + p) ∷_) ∘ rec $ begin-strict
      # t′                            ≡˘⟨ cong (#_ ∘ proj₂) eq ⟩
      # proj₂ (findDeleteMin t)       <⟨ n<1+n (# proj₂ (findDeleteMin t)) ⟩
      suc (# proj₂ (findDeleteMin t)) ≡⟨ #-findDeleteMin t ⟩
      # t                             ∎
      where
        -- I wish I could have instance patterns!
        instance _ = ne
        open ≤-Reasoning

    adjust : Heap → List (ℕ × ℕ) × Heap
    adjust = All.wfRec (On.wellFounded #_ <-wellFounded-fast) _ P adjust-rec
    
    adjusted : Heap
    adjusted with next , t′ ← adjust table′ = insert (p , nextComposite + p) (foldr insert t′ next)

sieve : List ℕ → List ℕ
sieve [] = []
sieve (x ∷ xs) = x ∷ sieve′ xs (insertPrime x empty)

primes : ℕ → List ℕ
primes n = sieve (drop 2 (upTo n))

